from biscuit.api import API
from biscuit.middleware import Middleware

APP = API()


class CustomMiddleware(Middleware):
    def process_request(self, request):
        print("Processing request", request.url)

    def process_response(self, request, response):
        print("Processing response", request.url)


APP.add_middleware(CustomMiddleware)


def custom_exc_handler(request, response, exception):
    response.text = str(exception)


APP.add_exception_handler(custom_exc_handler)


@APP.route("/home", methods=["get"])
def home(request, response):
    response.text = "Hi from home"


@APP.route("/about")
def about(request, response):
    response.text = "Hi from about"


@APP.route("/hello/{name}")
def welcome(request, response, name):
    response.text = f"hi {name}"


@APP.route("/article")
class ArticleView:
    def get(self, request, response):
        response.text = "Article GET"

    def post(self, request, response):
        response.text = "Article POST"


@APP.route("/exception")
def exception_throwing_handler(request, response):
    raise AssertionError("Won't happen")



import inspect
from parse import parse
from webob import Request, Response
from requests import Session as ReqSession
from wsgiadapter import WSGIAdapter as ReqWSGIAdapter

from .middleware import Middleware

DEFAULT_METHODS = ["get", "post", "put", "patch", "delete", "options"]


class API:
    """API for instantiating necessary PEP-333 call"""

    def __init__(self):
        self.routes = {}
        self.exception_handler = None
        self.middleware = Middleware(self)

    def __call__(self, environ, start_response):
        """PEP-333(3) use middleware as entrypoint"""
        return self.middleware(environ, start_response)

    def handle_request(self, request):
        """Handle a request and return correct path handler"""
        response = Response()

        handler_data, kwargs = self.select_handler(request.path)

        try:
            if handler_data is not None:
                handler = handler_data["handler"]
                allowed_methods = handler_data["allowed_methods"]
                if inspect.isclass(handler):
                    handler = getattr(handler(), request.method.lower(), None)
                    if handler is None:
                        raise AttributeError("Invalid method.", request.method)
                else:
                    if request.method.lower() not in allowed_methods:
                        raise AttributeError("Invalid method.", request.method)

                handler(request, response, **kwargs)
            else:
                self.default_response(response)
        except Exception as exc:
            if self.exception_handler is None:
                raise exc
            self.exception_handler(request, response, exc)

        return response

    def select_handler(self, req_path):
        """Select handler for request path"""
        for path, handler_data in self.routes.items():
            parse_result = parse(path, req_path)
            if parse_result is not None:
                return handler_data, parse_result.named

        return None, None

    def add_route(self, path, handler, allowed_methods):
        assert path not in self.routes, "Route already defined."

        if allowed_methods is None:
            allowed_methods = DEFAULT_METHODS
        self.routes[path] = {
            "handler": handler,
            "allowed_methods": allowed_methods,
        }

    def route(self, path, methods=None):
        """Handle routes"""

        def wrapper(handler):
            self.add_route(path, handler, methods)
            return handler

        return wrapper

    def default_response(self, response):
        """Return default response of 404 for paths not in self.routes"""
        response.status_code = 404
        response.text = "handler not found"

    def add_exception_handler(self, exception_handler):
        """Add custom exception handling"""
        self.exception_handler = exception_handler

    def add_middleware(self, middleware):
        """Add middleware"""
        self.middleware.add(middleware)

    def test_session(self, base_url="http://testserver"):
        """Test session for sending requests to the app"""
        session = ReqSession()
        session.mount(prefix=base_url, adapter=ReqWSGIAdapter(self))
        return session



from biscuit.api import API
import pytest
from biscuit.middleware import Middleware


APP = API()


class TestAPI:
    """Test API class"""

    def test_adding_route_handlers(self, mocker):
        """Should add a route handler"""
        test = APP.route("/test_route")

        test(mocker)

        assert APP.routes["/test_route"]

    def test_route_duplicate_throws_exception(self, api):
        @api.route("/duplicate")
        def duplicate(request, response):
            response.text = "test"

        with pytest.raises(AssertionError):

            @api.route("/duplicate")
            def duplicate2(request, response):
                response.text = "test"

    def test_custom_exception_handler(self, api, client):
        def on_exception(request, response, exc):
            response.text = "AttributeErrorHappened"

        api.add_exception_handler(on_exception)

        @api.route("/")
        def index(request, response):
            raise AttributeError()

        resp = client.get("http://testserver/")
        assert resp.text == "AttributeErrorHappened"

    def test_middleware(self, api, client):
        process_request_called = False
        process_response_called = False

        class CallMiddlewareMethods(Middleware):
            def __init__(self, app):
                super().__init__(app)

            def process_request(self, request):
                nonlocal process_request_called
                process_request_called = True

            def process_response(self, request, response):
                nonlocal process_response_called
                process_response_called = True

        api.add_middleware(CallMiddlewareMethods)

        @api.route("/")
        def index(request, response):
            response.text = "middleware"

        resp = client.get("http://testserver/")
        assert resp.text == "middleware"
        assert process_response_called is True
        assert process_request_called is True

    def test_allowed_methods(self, client, api):
        @api.route("/test", methods=["post"])
        def index(request, response):
            response.text = "Testing"

        with pytest.raises(AttributeError):
            client.get("http://testserver/test")

        assert client.post("http://testserver/test").text == "Testing"



from webob import Request


class Middleware:
    """Custom middleware"""

    def __init__(self, app):
        self.app = app

    def __call__(self, environ, start_response):
        """PEP 333 compliant entrypoint"""
        request = Request(environ)
        response = self.app.handle_request(request)
        return response(environ, start_response)

    def add(self, middleware):
        """Add middleware"""
        self.app = middleware(self.app)

    def process_request(self, request):
        """Process the request"""
        pass

    def process_response(self, request, response):
        """Process the response"""
        pass

    def handle_request(self, request):
        """Handle a request"""
        self.process_request(request)
        response = self.app.handle_request(request)
        self.process_response(request, response)
        return response



import requests
import pytest

from biscuit.api import API


class TestApp:
    """Test sample web application"""

    def test_default_response_when_no_path(self):
        """Should receive default response with no route handler"""
        resp = requests.get("http://localhost:8000/notreal")

        assert resp.text == "handler not found"
        assert resp.status_code == 404

    def test_handle_request_params(self):
        """Should handle parameterized request paths"""
        resp = requests.get("http://localhost:8000/hello/jake")

        assert resp.text == "hi jake"
        assert resp.status_code == 200

    def test_handle_request(self):
        """Should handle regular request paths"""
        resp = requests.get("http://localhost:8000/home")

        assert resp.text == "Hi from home"
        assert resp.status_code == 200

    def test_handle_class_based_view_get(self):
        """Should handle GET requests for CBV"""
        resp = requests.get("http://localhost:8000/article")

        assert resp.text == "Article GET"
        assert resp.status_code == 200

    def test_handle_class_based_view_post(self):
        """Should handle POST requests for CBV"""
        resp = requests.post("http://localhost:8000/article")

        assert resp.text == "Article POST"
        assert resp.status_code == 200



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import io
import os

from setuptools import find_packages, setup

# Package meta-data.
NAME = "biscuit"
DESCRIPTION = "Biscuit web framework"
EMAIL = "munsterberg@protonmail.ch"
AUTHOR = "munsterberg"
REQUIRES_PYTHON = ">=3.6.0"
VERSION = "0.0.1"

# Which packages are required for this module to be executed?
REQUIRED = [
    "parse==1.12.1",
    "requests==2.22.0",
    "requests-wsgi-adapter==0.4.1",
    "WebOb==1.8.5",
]

# What packages are optional?
EXTRAS = {
    # 'fancy feature': ['django'],
}

# The rest you shouldn't have to touch too much :)
# ------------------------------------------------
# Except, perhaps the License and Trove Classifiers!
# If you do change the License, remember to change the Trove Classifier for that!

here = os.path.abspath(os.path.dirname(__file__))

# Import the README and use it as the long-description.
# Note: this will only work if 'README.md' is present in your MANIFEST.in file!
try:
    with io.open(os.path.join(here, "README.md"), encoding="utf-8") as f:
        long_description = "\n" + f.read()
except FileNotFoundError:
    long_description = DESCRIPTION

# Load the package's __version__.py module as a dictionary.
about = {}
if not VERSION:
    project_slug = NAME.lower().replace("-", "_").replace(" ", "_")
    with open(os.path.join(here, project_slug, "__version__.py")) as f:
        exec(f.read(), about)
else:
    about["__version__"] = VERSION


# Where the magic happens:
setup(
    name=NAME,
    version=about["__version__"],
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type="text/markdown",
    author=AUTHOR,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    packages=find_packages(exclude=["test_*"]),
    install_requires=REQUIRED,
    include_package_data=True,
    license="MIT",
    classifiers=["Programming Language :: Python :: 3.6"],
    setup_requires=["wheel"],
)
